/*
 * Copyright (C) 2010 Jacob Kjome
 * 
 * This file is part of XMLC.
 *
 * XMLC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * XMLC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with XMLC.  If not,
 * see <http://www.gnu.org/licenses/>.
 */
package xmlc.demo;

import java.net.MalformedURLException;
import java.net.URL;
import java.util.HashSet;
import java.util.Set;

import org.w3c.dom.Attr;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.traversal.NodeFilter;

/**
 * Abstract NodeFilter for dealing with HTML Path related attributes.  Simply
 * implement the {@link #didModifyPathAttr(Set) didModifyPathAttr(Set&lt;Attr&gt;)}.
 * All other details are taken care of by this abstract class.
 * <p>
 * Path related attributes include "href", "src", "longdesc", "background",
 * "cite", "action", and "codebase".
 * </p>
 * <p>
 * Must iterate to apply changes.  Don't forget to detach to release resources!
 * </p>
 */
public abstract class AbstractPathModifierNodeFilter implements NodeFilter {
    private static final String[] HREF = {"href"},
        SRC_LONGDESC = {"src", "longdesc"},
        BACKGROUND = {"background"},
        CITE = {"cite"},
    //    PROFILE = {"profile"},
        ACTION = {"action"},
        CODEBASE = {"codebase"};

    /**
     * Would have filtered on ATTRIBUTE_NODE, but see
     * <a href="http://www.w3.org/TR/DOM-Level-2-Traversal-Range/traversal.html">DOM-Level-2-Traversal</a>
     * documentation below.  As such, filtering on ELEMENT_NODE and looking
     * up attributes of interest.
     * <h3>SHOW_ATTRIBUTE</h3>
     * <p>
     * <strong>Show Attr nodes</strong>. This is meaningful only when
     * creating an iterator or tree-walker with an attribute node as its
     * root; in this case, it means that the attribute node will appear in
     * the first position of the iteration or traversal. Since attributes
     * are never children of other nodes, they do not appear when traversing
     * over the document tree.
     * </p>
     */
    public final short acceptNode(final Node n) {
        if (n.getNodeType() == Node.ELEMENT_NODE) {
            final Element el = (Element) n;
            final Set/*<Attr>*/ attrNodeSet = new HashSet/*<Attr>*/(2);
            if (((hasAttrsThenPopulateAttrNodeSet(el, HREF, attrNodeSet)
//                && (el instanceof HTMLAnchorElement
//                    || el instanceof HTMLAreaElement
//                    || el instanceof HTMLBaseElement
//                    || el instanceof HTMLLinkElement)
                )
                || (hasAttrsThenPopulateAttrNodeSet(el, SRC_LONGDESC, attrNodeSet)
//                && (el instanceof HTMLScriptElement //no longdesc
//                    ||  el instanceof HTMLImageElement
//                    ||  el instanceof HTMLFrameElement
//                    ||  el instanceof HTMLIFrameElement
//                    ||  el instanceof HTMLInputElement) //no longdesc
                )
                || (hasAttrsThenPopulateAttrNodeSet(el, BACKGROUND, attrNodeSet)
//                && (el instanceof HTMLBodyElement)
                )
                || (hasAttrsThenPopulateAttrNodeSet(el, CITE, attrNodeSet)
//                && (el instanceof HTMLQuoteElement) //<BlockQuote>, <Q>, <Ins>, <Del>
                )
//                || (verifyAndPopulate(el, new String[] PROFILE, attrSet) //profile used as either globally unique name or a link - not safe to generically modify 
//                && (el instanceof HTMLHeadElement)
//                )
                || (hasAttrsThenPopulateAttrNodeSet(el, ACTION, attrNodeSet)
//                && (el instanceof HTMLFormElement)
                )
                || (hasAttrsThenPopulateAttrNodeSet(el, CODEBASE, attrNodeSet) //archive, classid, and data values all relative to codebase, or document (or <base href="...">) if codebase doesn't exist
//                && (el instanceof HTMLAppletElement
//                    || el instanceof HTMLObjectElement)
                ))
                && didModifyPathAttr(attrNodeSet)) {
                return NodeFilter.FILTER_ACCEPT;
            }
        }
        return NodeFilter.FILTER_SKIP;
    }

    /**
     * Called by the final {@link #acceptNode(Node)} method
     * 
     * @param pathAttrs a Set of Attr nodes, for a given element, that have been
     *                  identified as representing a URI path, may be empty but
     *                  never null
     * @return true to indicate at least one Attr value was modified, otherwise
     *         false
     */
    protected abstract boolean didModifyPathAttr(Set/*<Attr>*/ pathAttrs);

    /**
     * Checks if the path is fully qualified.  Note that null/empty paths (same
     * goes for paths starting with '?') are defacto fully qualified because
     * this generally means that the link is supposed to point to the current
     * page.  Qualifying the path in this case would break the link.  Similarly,
     * if the path is pre-qualified (i.e., starts with '/'), qualifying it further
     * would break the link.  More obviously, if the path defines a protocol,
     * any modification of the path would break the link.
     * 
     * @param path the URI path (null safe)
     * @return true if there is no path at all (null or empty), the path starts
     *         with a non-alphanumeric character (presumably a '/' or '?'), or
     *         the path defines a protocol, otherwise false
     */
    public boolean isFullQualifiedPath(final String path) {
        return !hasURLParts(path) || getURLPart(path, 0) != null;
    }

    /**
     * @param path the URI path (null safe)
     * @return the host part if it exists, otherwise null
     */
    public String getHost(final String path) {
        return hasURLParts(path) ? getURLPart(path, 1) : null;
    }

    /**
     * @param path the URI path (null safe)
     * @return the path part if it exists, otherwise null
     */
    public String getPath(final String path) {
        return hasURLParts(path) ? getURLPart(path, 2) : null;
    }

    private static boolean hasURLParts(final String path) {
        return !(path == null || path.length() == 0 || !Character.isLetterOrDigit(path.charAt(0)));
    }

    private static String getURLPart(final String path, int part) {
        try {
            final URL url = new URL(path);
            switch(part) {
                case 0:
                    return url.getProtocol();
                case 1:
                    return url.getHost();
                case 2:
                    return url.getPath();
                default:
                    return null;
            }            
        } catch (final MalformedURLException mue) {
            return null;
        }
    }

    private static boolean hasAttrsThenPopulateAttrNodeSet(final Element el, final String[] attrNames, final Set/*<Attr>*/ attrNodeSet) {
        boolean retVal = false;
        for (int i = 0; i < attrNames.length; i++) {
            final Attr attr = el.getAttributeNode(attrNames[i]);
            if (attr != null && attrNodeSet.add(attr)) {
                retVal = true;
            }
        }
        return retVal;
    }

}
