/*
 * Copyright (C) 2010 Jacob Kjome
 * 
 * This file is part of XMLC.
 *
 * XMLC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * XMLC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with XMLC.  If not,
 * see <http://www.gnu.org/licenses/>.
 */
package xmlc.demo;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.TreeMap;

import org.enhydra.xml.dom.DOMOps;
import org.w3c.dom.Document;
import org.w3c.dom.DocumentFragment;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.ranges.DocumentRange;
import org.w3c.dom.ranges.Range;

/**
 * A class making it easy to decorate a DOM with Strings and Nodes.  Nodes are
 * adopted as needed relieving the burden of manually adopting nodes.  Advanced
 * DOM techniques, such as DOM Level 2 Document Ranges, are made possible without
 * exposing complexity to the user.
 */
public class DOMDecorator {
    private final Map/*<String, Node>*/ sectionNewParentNodes = new HashMap/*<String, Node>*/();
    private final Map/*<String, NodeInfo>*/ sectionReplaceWithNodes = new HashMap/*<String, NodeInfo>*/()
        , sectionPrependWithNodes = new TreeMap/*<String, NodeInfo>*/()
        , sectionAppendWithNodes = new TreeMap/*<String, NodeInfo>*/();
    private final Map/*<String, String>*/ sectionReplaceWithStrings = new HashMap/*<String, String>*/();

    private static class NodeInfo {
        final Node node;
        final boolean extractNodeContents;

        NodeInfo(final Node node, final boolean extractNodeContents) {
            this.node = node;
            this.extractNodeContents = extractNodeContents;
        }
    }

    public DOMDecorator surroundSectionWith(final String sectionId, final Node newParentNode) {
        if (sectionId == null || newParentNode == null) return this;
        this.sectionNewParentNodes.put(sectionId, newParentNode);
        return this;
    }

    public DOMDecorator prependChildSectionWith(final String sectionId, final Node node) {
        if (sectionId == null || node == null) return this;
        this.sectionPrependWithNodes.put(positionKey(sectionId, sectionPrependWithNodes.size()), new NodeInfo(node, false));
        return this;
    }

    public DOMDecorator prependChildSectionWithContents(final String sectionId, final Node contentContainerNode) {
        if (sectionId == null || contentContainerNode == null) return this;
        this.sectionPrependWithNodes.put(positionKey(sectionId, sectionPrependWithNodes.size()), new NodeInfo(contentContainerNode, true));
        return this;
    }

    public DOMDecorator appendChildSectionWith(final String sectionId, final Node node) {
        if (sectionId == null || node == null) return this;
        this.sectionAppendWithNodes.put(positionKey(sectionId, sectionAppendWithNodes.size()), new NodeInfo(node, false));
        return this;
    }

    public DOMDecorator appendChildSectionWithContents(final String sectionId, final Node contentContainerNode) {
        if (sectionId == null || contentContainerNode == null) return this;
        this.sectionAppendWithNodes.put(positionKey(sectionId, sectionAppendWithNodes.size()), new NodeInfo(contentContainerNode, true));
        return this;
    }

    public DOMDecorator clearSection(final String sectionId) {
        return replaceSectionWith(sectionId, (String) null);
    }

    public DOMDecorator replaceSectionWith(final String sectionId, final String sectionValue) {
        if (sectionId == null) return this;
        this.sectionReplaceWithStrings.put(sectionId, sectionValue);
        return this;
    }

    public DOMDecorator replaceSectionWith(final String sectionId, final Node node) {
        return replaceSectionWithNodeOrContentsImpl(sectionId, node, false);
    }

    public DOMDecorator replaceSectionWithContents(final String sectionId, final Node contentContainerNode) {
        return replaceSectionWithNodeOrContentsImpl(sectionId, contentContainerNode, true);
    }

    private DOMDecorator replaceSectionWithNodeOrContentsImpl(final String sectionId, final Node node, final boolean extractNodeContents) {
        if (sectionId == null) return this;
        if (node == null) {
            return clearSection(sectionId);
        }
        sectionReplaceWithNodes.put(sectionId, new NodeInfo(node, extractNodeContents));
        return this;
    }

    public DOMDecorator reset() {
        sectionNewParentNodes.clear();
        sectionReplaceWithNodes.clear();
        sectionPrependWithNodes.clear();
        sectionAppendWithNodes.clear();
        sectionReplaceWithStrings.clear();
        return this;
    }

    public boolean decorate(Document decoratorDoc) {
        boolean modified = false;
        if (decoratorDoc == null) return modified;        
        decoratorDoc = DOMOps.getDocument(decoratorDoc);

        for (final Iterator/*<Map.Entry<String, Node>>*/ iter = sectionNewParentNodes.entrySet().iterator(); iter.hasNext();) {
            final Map.Entry/*<String, Node>*/ entry = (Map.Entry) iter.next();
            final String sectionId = (String) entry.getKey();
            final Node sectionNewParentNode = (Node) entry.getValue();
            
            final Range targetRange = prepareTargetRangeAndNewParentNodeForSurrounding(decoratorDoc, sectionId, sectionNewParentNode);
            if (targetRange == null) continue;
            targetRange.surroundContents(sectionNewParentNode);
            modified = true;
            targetRange.detach();
        }

        for (final Iterator/*<Map.Entry<String, NodeInfo>>*/ iter = sectionReplaceWithNodes.entrySet().iterator(); iter.hasNext();) {
            final Map.Entry/*<String, NodeInfo>*/ entry = (Map.Entry) iter.next();
            final String sectionId = (String) entry.getKey();
            final NodeInfo sectionNodeInfo = (NodeInfo) entry.getValue();
            
            final Range targetRange = prepareTargetRangeForReplacement(decoratorDoc, sectionId);
            if (targetRange == null) continue;
            modified = true;
            targetRange.insertNode(prepareNode(decoratorDoc, sectionNodeInfo));
            targetRange.detach();
        }

        if (!sectionReplaceWithStrings.isEmpty()) {
            sectionReplaceWithStrings.keySet().removeAll(sectionReplaceWithNodes.keySet());
            for (final Iterator/*<Map.Entry<String, String>>*/ iter = sectionReplaceWithStrings.entrySet().iterator(); iter.hasNext();) {
                final Map.Entry/*<String, String>*/ entry = (Map.Entry) iter.next();
                final String sectionId = (String) entry.getKey();
                final String sectionString = (String) entry.getValue();

                final Range targetRange = prepareTargetRangeForReplacement(decoratorDoc, sectionId);
                if (targetRange == null) continue;
                modified = true;
                if (sectionString != null) {
                    targetRange.insertNode(decoratorDoc.createTextNode(sectionString));
                }
                targetRange.detach();
            }
        }

        if (prependOrAppendNodeOrContents(decoratorDoc, sectionPrependWithNodes, true)) {
            modified = true;
        }

        if (prependOrAppendNodeOrContents(decoratorDoc, sectionAppendWithNodes, false)) {
            modified = true;
        }

        return modified;
    }

    private static boolean prependOrAppendNodeOrContents(final Document decoratorDoc, final Map/*<String, NodeInfo>*/ sectionNodes, final boolean prepend) {
        boolean modified = false;
        for (final Iterator/*<Map.Entry<String, NodeInfo>>*/ iter = sectionNodes.entrySet().iterator(); iter.hasNext();) {
            final Map.Entry/*<String, NodeInfo>*/ entry = (Map.Entry) iter.next();
            final String sectionId = dePositionKey((String) entry.getKey());
            final NodeInfo sectionNodeInfo = (NodeInfo) entry.getValue();

            final Element targetEl = decoratorDoc.getElementById(sectionId);
            if (targetEl == null) continue;
            final DocumentFragment df = prepareNode(decoratorDoc, sectionNodeInfo);
            final Node firstChild;
            if (prepend && (firstChild = targetEl.getFirstChild()) != null) {
                targetEl.insertBefore(df, firstChild);
            } else {
                targetEl.appendChild(df);
            }
            modified = true;
        }
        return modified;
    }

    private static Range prepareTargetRangeAndNewParentNodeForSurrounding(final Document decoratorDoc, final String sectionId, final Node sectionNewParentNode) {
        final Element targetEl = decoratorDoc.getElementById(sectionId);
        if (targetEl == null) return null;
        final Range targetRange = ((DocumentRange) decoratorDoc).createRange();
        targetRange.selectNode(targetEl);
        if (!decoratorDoc.isSameNode(DOMOps.getDocument(sectionNewParentNode))) {
            decoratorDoc.adoptNode(sectionNewParentNode);
        }
        return targetRange;
    }

    private static Range prepareTargetRangeForReplacement(final Document decoratorDoc, final String sectionId) {
        final Element targetEl = decoratorDoc.getElementById(sectionId);
        if (targetEl == null) return null;
        final Range targetRange = ((DocumentRange) decoratorDoc).createRange();
        targetRange.setStartBefore(targetEl.getFirstChild());
        targetRange.setEndAfter(targetEl.getLastChild());
        targetRange.deleteContents();
        return targetRange;
    }

    /**
     * Same as
     * {@link #prepareNode(Document, Node, boolean) prepareNode(targetDoc, node, false)}
     * 
     * @param targetDoc the target document for node inclusion 
     * @param node the node to prepare for inclusion in the target document
     * @return a document fragment representing the provided node, prepared for
     *         inclusion in the target document
     * @see #prepareNode(Document, Node, boolean)
     */
    public static DocumentFragment prepareNode(final Document targetDoc, final Node node) {
        return prepareNode(targetDoc, node, false);
    }

    /**
     * Prepares the provided node, or its contents, for inclusion in the provided
     * target document.  If the provided node is not already a member of the
     * target document, it will be adopted.
     * 
     * @param targetDoc the target document for node inclusion 
     * @param node the node to prepare for inclusion in the target document
     * @param extractNodeContents whether to extract and prepare the node contents
     *                            rather than the node itself
     * @return a document fragment representing the provided node, prepared for
     *         inclusion in the target document, an empty document fragment if
     *         targetDoc is non-null and node is null, or null if targetDoc is null
     */
    public static DocumentFragment prepareNode(final Document targetDoc, final Node node, final boolean extractNodeContents) {
        if (targetDoc != null && node == null) return targetDoc.createDocumentFragment();
        else if (targetDoc == null || node == null) return null;
        return prepareNode(targetDoc, new NodeInfo(node, extractNodeContents));
    }

    private static DocumentFragment prepareNode(final Document targetDoc, final NodeInfo nodeInfo) {
        final Node n = nodeInfo.node;
        final Document d = DOMOps.getDocument(n);
        final DocumentFragment df;
        if (DOMOps.isNodeOfType(n, Node.DOCUMENT_FRAGMENT_NODE)) {
            df = (DocumentFragment) n;
        } else if (!nodeInfo.extractNodeContents || n.getFirstChild() == null) {
            df = d.createDocumentFragment();
            df.appendChild(n);
        } else {
            final Range sectionRange = ((DocumentRange) d).createRange();
            if (n.getParentNode() == null) {
                d.createDocumentFragment().appendChild(n);
            }
            sectionRange.setStartBefore(n.getFirstChild());
            sectionRange.setEndAfter(n.getLastChild());
            df = sectionRange.extractContents();
            sectionRange.detach();
        }
        if (!targetDoc.isSameNode(d)) {
            targetDoc.adoptNode(df);
        }
        return df;
    }

    /*
     * Used to make unique and get JDK 1.4.2+ LinkedHashMap insertion order
     * behavior from a TreeMap.
     */
    private static String positionKey(final String key, final int position) {
        final String positionStr = Integer.toString(position);
        final StringBuffer b = new StringBuffer(key.length() + positionStr.length() + 1);
        return b.append(positionStr).append(']').append(key).toString();
    }

    /*
     * Used to obtain the original key, less the positioning information
     */
    private static String dePositionKey(String positionedkey) {
        return positionedkey.substring(positionedkey.indexOf(']') + 1);
    }

    public static DOMDecoratorContext newEl(Document doc, String localName) {
        return newEl(doc, localName, null);
    }

    public static DOMDecoratorContext newEl(Document doc, String localName, String namespace) {
        if (doc == null || localName == null) return null;
        final Element el = namespace != null ? doc.createElementNS(localName, namespace) : doc.createElement(localName);
        return new DOMDecoratorContext(el);
    }

    public static class DOMDecoratorContext {
        final Element el;

        private DOMDecoratorContext(final Element el) {
            if (el == null) {
                throw new IllegalArgumentException("Argument 'el' must not be null");
            }
            this.el = el;
        }

        public DOMDecoratorContext setTextContent(String value) {
            if (value != null) el.setTextContent(value);
            return this;
        }

        public DOMDecoratorContext setAttribute(String name, String value) {
            if (name != null && value != null) el.setAttribute(name, value);
            return this;
        }

        public DOMDecoratorContext appendChild(final DOMDecoratorContext context) {
            if (context != null) return appendChild(context.el);
            return this;
        }

        public DOMDecoratorContext appendChild(final Node node) {
            if (node != null) el.appendChild(node);
            return this;
        }

        public Element getElement() {
            return el;
        }
    }

}
